/*
This file is part of libmspcore     
Copyright © 2006  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#include "timer.h"
#include "utils.h"

using namespace std;

namespace Msp {
namespace Time {

Timer::~Timer()
{
	for(set<SlotProxy>::iterator i=slots.begin(); i!=slots.end(); ++i)
		delete i->slot;
}

Timer::Slot &Timer::add(const TimeDelta &td)
{
	Slot *s=new Slot(td);
	mutex.lock();
	slots.insert(s);
	mutex.unlock();
	sem.signal();
	return *s;
}

Timer::Slot &Timer::add(const TimeStamp &ts)
{
	Slot *s=new Slot(ts);
	{
		MutexLock l(mutex);
		slots.insert(s);
	}
	sem.signal();
	return *s;
}

void Timer::cancel(Slot &slot)
{
	MutexLock l(mutex);
	if(slots.erase(&slot))
		delete &slot;
}

void Timer::tick(bool block)
{
	if(slots.empty())
	{
		if(block)
			sem.wait();
		else
			return;
	}

	Slot *next;
	{
		MutexLock l(mutex);
		next=slots.begin()->slot;
	}

	const TimeStamp &stamp=next->get_timeout();
	const TimeStamp t=now();
	if(stamp<=t || (block && sem.wait(stamp-t)==1))
	{
		slots.erase(slots.begin());
		if(next->signal_timeout.emit() && next->increment())
			slots.insert(next);
		else
			delete next;
	}
}

TimeStamp Timer::get_next_timeout() const
{
	if(slots.empty())
		return TimeStamp();
	return slots.begin()->slot->get_timeout();
}


Timer::Slot::Slot(const TimeDelta &td):
	interval(td),
	timeout(now()+interval)
{ }

Timer::Slot::Slot(const TimeStamp &ts):
	timeout(ts)
{ }

bool Timer::Slot::increment()
{
	if(!interval)
		return false;
	timeout+=interval;
	return true;
}


Timer::SlotProxy::SlotProxy(Slot *s):
	slot(s)
{ }

bool Timer::SlotProxy::operator<(const SlotProxy &sp) const
{
	return slot->get_timeout()<sp.slot->get_timeout();
}

} // namespace Time
} // namespace Msp
